from abc import ABC, abstractmethod
from typing import Optional, Sequence

from qiskit.providers import BackendV2


class BaseProvider(ABC):
    """Base class for a provider."""

    @abstractmethod
    def get_backend(self, name: Optional[str] = None, id: Optional[int] = None) -> BackendV2:
        """Get a backend by name."""
        pass

    @abstractmethod
    def backends(self) -> Sequence[BackendV2]:
        """Return all backends for this provider."""
        pass



from datetime import datetime
from functools import cache
from typing import Any, List, Union, cast

from compute_api_client import Result as RawJobResult
from qi2_shared.hybrid.quantum_interface import QuantumInterface
from qiskit import QuantumCircuit, transpile
from qiskit.providers.backend import BackendV2
from qiskit.providers.jobstatus import JobStatus
from qiskit.result.result import Result

from qiskit_quantuminspire import cqasm
from qiskit_quantuminspire.qi_jobs import QIBaseJob


class QIHybridJob(QIBaseJob):
    """Used internally as a Qiskit job for hybrid algorithms that are fully executed on the Quantum Inspire platform.

    Not to be used in user code.
    """

    def __init__(
        self,
        run_input: Union[QuantumCircuit, List[QuantumCircuit]],
        backend: Union[BackendV2, None],
        quantum_interface: QuantumInterface,
        **kwargs: Any,
    ) -> None:
        self._quantum_interface = quantum_interface
        super().__init__(run_input, backend, **kwargs)

    def status(self) -> JobStatus:
        """Quantum jobs in a hybrid environment are executed synchronously."""
        return JobStatus.DONE

    def submit(self) -> None:
        options = cast(dict[str, Any], self.backend().options)

        for circuit_data in self.circuits_run_data:
            transpiled_circuit = transpile(circuit_data.circuit, backend=self.backend())
            result = self._quantum_interface.execute_circuit(
                cqasm.dumps(transpiled_circuit), options.get("shots"), raw_data_enabled=options.get("memory")
            )

            # Store retrieved results in format expected by Job
            circuit_data.results = RawJobResult(
                id=0,
                created_on=datetime.now(),
                job_id=0,
                metadata_id=0,
                execution_time_in_seconds=0,
                shots_requested=result.shots_requested,
                shots_done=result.shots_done,
                results=result.results,
                raw_data=result.raw_data,
            )

    @cache
    def result(self) -> Result:
        return self._process_results()



import asyncio
import concurrent
from typing import Any, Coroutine

from qiskit.transpiler import CouplingMap


def is_coupling_map_complete(coupling_map: CouplingMap) -> bool:
    """A complete digraph is a digraph in which there is a directed edge from every vertex to every other vertex."""
    distance_matrix = coupling_map.distance_matrix

    assert distance_matrix is not None

    is_semicomplete = all(distance in [1, 0] for distance in distance_matrix.flatten())

    return is_semicomplete and coupling_map.is_symmetric


def run_async(async_function: Coroutine[Any, Any, Any]) -> Any:
    try:
        _ = asyncio.get_running_loop()
        with concurrent.futures.ThreadPoolExecutor() as executor:
            return executor.submit(asyncio.run, async_function).result()
    except RuntimeError:
        return asyncio.run(async_function)



from typing import Any

from opensquirrel import CircuitBuilder
from opensquirrel.ir import Bit, Qubit
from opensquirrel.writer import writer
from qiskit import QuantumCircuit
from qiskit.circuit import CircuitInstruction

from qiskit_quantuminspire.mapping.instruction_mapping import InstructionMapping

_INSTRUCTION_MAPPING = InstructionMapping()


def _add_instruction(builder: CircuitBuilder, circuit_instruction: Any) -> None:
    operation = circuit_instruction.operation
    name = operation.name
    params = [param for param in operation.params]
    qubit_operands = [Qubit(qubit._index) for qubit in circuit_instruction.qubits]
    clbit_operands = [Bit(clbit._index) for clbit in circuit_instruction.clbits]

    try:
        # Get the gate's method in the CircuitBuilder class, call with operands
        # All of the builder's methods follow the same pattern, first the qubit operands, then parameters
        # Only method with classical bit operands is measure, which does not have parameters
        getattr(builder, _INSTRUCTION_MAPPING.qiskit_to_opensquirrel(name))(*qubit_operands, *clbit_operands, *params)
    except KeyError:
        raise NotImplementedError(
            f"Unsupported instruction: {name}. Please edit your circuit or use Qiskit transpilation to support "
            + "your selected backend."
        )


def dumps(circuit: QuantumCircuit) -> str:
    """Return the cQASM representation of the circuit."""
    builder = CircuitBuilder(circuit.num_qubits, circuit.num_clbits)
    for circuit_instruction in circuit.data:
        operation = circuit_instruction.operation
        name = operation.name

        if name == "delay":
            if circuit_instruction.operation.unit != "dt":
                raise NotImplementedError(
                    f"Unsupported delay unit {circuit_instruction.operation.unit} in: {circuit_instruction}. Only 'dt'"
                    + " is supported."
                )

        if name == "barrier":
            # Opensquirrel does not support multi-qubit barriers.
            for qubit in circuit_instruction.qubits:
                _add_instruction(builder, CircuitInstruction(operation=operation, qubits=[qubit]))
        else:
            _add_instruction(builder, circuit_instruction)

    cqasm: str = writer.circuit_to_string(builder.to_circuit())

    return cqasm



from typing import Any, List, Optional, Sequence

from compute_api_client import ApiClient, BackendType, BackendTypesApi, PageBackendType
from qi2_shared.client import config
from qi2_shared.pagination import PageReader

from qiskit_quantuminspire.base_provider import BaseProvider
from qiskit_quantuminspire.qi_backend import QIBackend
from qiskit_quantuminspire.utils import run_async


class QIProvider(BaseProvider):
    """List QIBackends integrated with QiskitBackend interface."""

    def __init__(self) -> None:
        self._qiskit_backends = self._construct_backends()

    async def _fetch_qi_backend_types(self) -> List[BackendType]:
        """Fetch backend types from CJM using api client.

        (Implemented without paging only for demonstration purposes, should get a proper implementation)
        """
        async with ApiClient(config()) as client:
            page_reader = PageReader[PageBackendType, BackendType]()
            backend_types_api = BackendTypesApi(client)
            backend_types: List[BackendType] = await page_reader.get_all(
                backend_types_api.read_backend_types_backend_types_get
            )
        return backend_types

    def _construct_backends(self) -> List[QIBackend]:
        """Construct QIBackend using fetched backendtypes and metadata."""
        qi_backend_types = run_async(self._fetch_qi_backend_types())
        qi_backends = [QIBackend(provider=self, backend_type=backend_type) for backend_type in qi_backend_types]
        return qi_backends

    def backends(self) -> Sequence[QIBackend]:
        return self._qiskit_backends

    def get_backend(self, name: Optional[str] = None, id: Optional[int] = None) -> QIBackend:
        filter_arguments: dict[str, Any] = {}

        if name is not None:
            filter_arguments["name"] = name

        if id is not None:
            filter_arguments["id"] = id

        for backend in self._qiskit_backends:
            if all(getattr(backend, key) == value for key, value in filter_arguments.items()):
                return backend

        raise ValueError(f"Backend {name} not found")



from typing import Any, List, Union

from compute_api_client import BackendType
from qi2_shared.hybrid.quantum_interface import QuantumInterface
from qiskit import QuantumCircuit
from qiskit.providers.models.backendstatus import BackendStatus

from qiskit_quantuminspire.hybrid.hybrid_job import QIHybridJob
from qiskit_quantuminspire.qi_backend import QIBaseBackend


class QIHybridBackend(QIBaseBackend):
    """Used as a Qiskit backend for hybrid algorithms that are fully executed on the Quantum Inspire platform.

    Quantum hardware specifications are inferred from the backend type selected on submission.
    """

    def __init__(self, qi: QuantumInterface, **kwargs: Any):
        super().__init__(BackendType.model_validate(qi.backend_type), **kwargs)
        self._quantum_interface = qi

    @property
    def status(self) -> BackendStatus:
        """Return the backend status. Pending jobs is always 0. This information is currently not known.

        Returns:
            BackendStatus: the status of the backend. Pending jobs is always 0.
        """
        return BackendStatus(
            backend_name=self._quantum_interface.backend_type.get("name", "Name unknown"),
            backend_version="2.0",
            operational=True,
            pending_jobs=0,
            status_msg="online",
        )

    def run(self, run_input: Union[QuantumCircuit, List[QuantumCircuit]], **options: Any) -> QIHybridJob:
        self.set_options(**options)
        job = QIHybridJob(run_input=run_input, backend=self, quantum_interface=self._quantum_interface)
        job.submit()
        return job



import asyncio
import warnings
from dataclasses import dataclass
from functools import cache
from pathlib import Path
from typing import Any, List, Optional, Union, cast

from colorama import Fore, Style
from compute_api_client import (
    Algorithm,
    AlgorithmIn,
    AlgorithmsApi,
    AlgorithmType,
    ApiClient,
    BatchJob,
    BatchJobIn,
    BatchJobsApi,
    BatchJobStatus,
    Commit,
    CommitIn,
    CommitsApi,
    CompileStage,
    File,
    FileIn,
    FilesApi,
    Job,
    JobIn,
    JobsApi,
    JobStatus as QIJobStatus,
    Language,
    LanguagesApi,
    PageBatchJob,
    PageResult,
    Project,
    ProjectIn,
    ProjectsApi,
    Result as RawJobResult,
    ResultsApi,
    ShareType,
)
from qi2_shared.client import config
from qi2_shared.pagination import PageReader
from qi2_shared.settings import ApiSettings
from qiskit import qpy
from qiskit.circuit import QuantumCircuit
from qiskit.providers import JobV1
from qiskit.providers.backend import BackendV2
from qiskit.providers.jobstatus import JobStatus
from qiskit.qobj import QobjExperimentHeader
from qiskit.result.models import ExperimentResult, ExperimentResultData
from qiskit.result.result import Result

from qiskit_quantuminspire import cqasm
from qiskit_quantuminspire.base_provider import BaseProvider
from qiskit_quantuminspire.utils import run_async


class ExperimentFailedWarning(UserWarning):
    pass


@dataclass
class CircuitExecutionData:
    """Class for book-keeping of individual jobs."""

    circuit: QuantumCircuit
    job_id: Optional[int] = None
    results: Optional[RawJobResult] = None
    system_message: Optional[str] = None


# Ignore type checking for QIBaseJob due to missing Qiskit type stubs,
# which causes the base class 'Job' to be treated as 'Any'.
class QIBaseJob(JobV1):  # type: ignore[misc]
    circuits_run_data: List[CircuitExecutionData]

    def __init__(
        self,
        run_input: Union[QuantumCircuit, List[QuantumCircuit]],
        backend: Union[BackendV2, None],
        **kwargs: Any,
    ) -> None:
        """Initialize a QIJob instance.

        Args:
            run_input: A single/list of Qiskit QuantumCircuit object(s).
            backend: The backend on which the job is run. While specified as `Backend` to avoid
                circular dependency, it is a `QIBackend`.
            **kwargs: Additional keyword arguments passed to the parent `Job` class.
        """
        super().__init__(backend, "", **kwargs)
        self.circuits_run_data = []
        self._add_circuits(run_input)
        self.program_name = "Program created by SDK"
        self.batch_job_id: Union[int, None] = None

    def _add_circuits(self, circuits: Union[QuantumCircuit, List[QuantumCircuit]]) -> None:
        """Add circuits to the list of circuits to be run."""
        circuits = [circuits] if isinstance(circuits, QuantumCircuit) else circuits
        self.circuits_run_data.extend([CircuitExecutionData(circuit=circuit) for circuit in circuits])

    def _process_results(self) -> Result:
        """Process the raw job results obtained from QuantumInspire."""

        results = []
        batch_job_success = [False] * len(self.circuits_run_data)
        failed_experiments = dict()

        for idx, circuit_data in enumerate(self.circuits_run_data):
            qi_result = circuit_data.results
            circuit_name = circuit_data.circuit.name

            if qi_result is None:
                failed_experiments[circuit_name] = circuit_data.system_message
                experiment_result = self._create_empty_experiment_result(
                    circuit_name=circuit_name, system_message=circuit_data.system_message
                )
                results.append(experiment_result)
                continue

            experiment_result = self._create_experiment_result(
                circuit_name=circuit_name,
                result=qi_result,
            )
            results.append(experiment_result)
            batch_job_success[idx] = qi_result.shots_done > 0

        if failed_experiments:

            warning_message = (
                f"\n{Fore.YELLOW}Some experiments {Fore.RED}FAILED. "
                f"{Fore.YELLOW}You can view the detailed system messages \n"
                f"in the 'system_messages' attribute of the result object. "
                f"\nFor e.g: result.system_messages.{Style.RESET_ALL}"
            )
            warnings.warn(warning_message, category=ExperimentFailedWarning)

        result = Result(
            backend_name=self.backend().name,
            backend_version="1.0.0",
            qobj_id="",
            job_id=str(self.batch_job_id),
            success=all(batch_job_success),
            results=results,
            status="Result successful" if all(batch_job_success) else "Result failed",
            system_messages=failed_experiments,
        )
        return result

    @staticmethod
    def _create_experiment_result(
        circuit_name: str,
        result: RawJobResult,
    ) -> ExperimentResult:
        """Create an ExperimentResult instance based on RawJobResult parameters."""
        counts = {hex(int(key, 2)): value for key, value in result.results.items()}
        memory = [hex(int(measurement, 2)) for measurement in result.raw_data] if result.raw_data else None

        experiment_data = ExperimentResultData(
            counts={} if counts is None else counts,
            memory=memory,
        )
        return ExperimentResult(
            shots=result.shots_done,
            success=result.shots_done > 0,
            data=experiment_data,
            header=QobjExperimentHeader(name=circuit_name),
            status="Experiment successful",
        )

    @staticmethod
    def _create_empty_experiment_result(circuit_name: str, system_message: Optional[str]) -> ExperimentResult:
        """Create an empty ExperimentResult instance."""
        return ExperimentResult(
            shots=0,
            success=False,
            data=ExperimentResultData(counts={}),
            header=QobjExperimentHeader(name=circuit_name),
            status=f"Experiment failed. System Message: {system_message}",
        )


class QIJob(QIBaseJob):
    """A wrapper class for QuantumInspire batch jobs to integrate with Qiskit's Job interface."""

    def submit(self) -> None:
        run_async(self._submit_async())

    async def _submit_async(self) -> None:
        """Submit the (batch)job to the quantum inspire backend.

        Use compute-api-client to call the cjm endpoints in the correct order, to submit the jobs.
        """
        options = cast(dict[str, Any], self.backend().options)
        configuration = config()
        settings = ApiSettings.from_config_file()

        # call create algorithm
        async with ApiClient(configuration) as api_client:
            language = await self._get_language(api_client, "cqasm", "3.0")
            if language is None:
                raise RuntimeError("No cqasm v3.0 language id returned by the platform")

            team_member_id = settings.auths[settings.default_host].team_member_id
            assert isinstance(team_member_id, int)

            project = await self._create_project(api_client, team_member_id)
            batch_job = await self._create_batch_job(api_client, backend_type_id=self.backend().id)

            async def job_run_sequence(
                in_api_client: ApiClient,
                in_project: Project,
                in_batch_job: BatchJob,
                circuit_data: CircuitExecutionData,
            ) -> None:
                algorithm = await self._create_algorithm(in_api_client, in_project.id)
                commit = await self._create_commit(in_api_client, algorithm.id)
                file = await self._create_file(in_api_client, commit.id, language.id, circuit_data.circuit)
                job: Job = await self._create_job(
                    in_api_client,
                    file.id,
                    in_batch_job.id,
                    raw_data_enabled=cast(bool, options.get("memory")),
                    number_of_shots=options.get("shots"),
                )
                circuit_data.job_id = job.id

            # iterate over the circuits
            run_coroutines = (
                job_run_sequence(api_client, project, batch_job, circuit_run_data)
                for circuit_run_data in self.circuits_run_data
            )
            await asyncio.gather(*run_coroutines)
            await self._enqueue_batch_job(api_client, batch_job.id)
            self.batch_job_id = batch_job.id

    async def _create_project(self, api_client: ApiClient, owner_id: int) -> Project:
        api_instance = ProjectsApi(api_client)
        obj = ProjectIn(
            owner_id=owner_id,
            name=self.program_name,
            description="Project created by SDK",
            starred=False,
        )
        return await api_instance.create_project_projects_post(obj)

    async def _create_algorithm(self, api_client: ApiClient, project_id: int) -> Algorithm:
        api_instance = AlgorithmsApi(api_client)
        obj = AlgorithmIn(
            project_id=project_id, type=AlgorithmType.QUANTUM, shared=ShareType.PRIVATE, name=self.program_name
        )
        return await api_instance.create_algorithm_algorithms_post(obj)

    async def _create_commit(self, api_client: ApiClient, algorithm_id: int) -> Commit:
        api_instance = CommitsApi(api_client)
        obj = CommitIn(
            description="Commit created by SDK",
            algorithm_id=algorithm_id,
        )
        return await api_instance.create_commit_commits_post(obj)

    async def _create_file(
        self, api_client: ApiClient, commit_id: int, language_id: int, circuit: QuantumCircuit
    ) -> File:
        api_instance = FilesApi(api_client)
        obj = FileIn(
            commit_id=commit_id,
            content=cqasm.dumps(circuit),
            language_id=language_id,
            compile_stage=CompileStage.NONE,
            compile_properties={},
        )
        return await api_instance.create_file_files_post(obj)

    async def _create_batch_job(self, api_client: ApiClient, backend_type_id: int) -> BatchJob:
        api_instance = BatchJobsApi(api_client)
        obj = BatchJobIn(backend_type_id=backend_type_id)
        return await api_instance.create_batch_job_batch_jobs_post(obj)

    async def _create_job(
        self,
        api_client: ApiClient,
        file_id: int,
        batch_job_id: int,
        raw_data_enabled: bool,
        number_of_shots: Optional[int] = None,
    ) -> Job:
        api_instance = JobsApi(api_client)
        obj = JobIn(
            file_id=file_id,
            batch_job_id=batch_job_id,
            number_of_shots=number_of_shots,
            raw_data_enabled=raw_data_enabled,
        )
        return await api_instance.create_job_jobs_post(obj)

    async def _enqueue_batch_job(self, api_client: ApiClient, batch_job_id: int) -> BatchJob:
        api_instance = BatchJobsApi(api_client)
        return await api_instance.enqueue_batch_job_batch_jobs_id_enqueue_patch(batch_job_id)

    async def _get_language(
        self, api_client: ApiClient, language_name: str, language_version: str
    ) -> Union[Language, None]:
        language_api_instance = LanguagesApi(api_client)
        languages_page = await language_api_instance.read_languages_languages_get()
        for lan in languages_page.items:
            if language_name.lower() == lan.name.lower():
                if language_version == lan.version:
                    return lan

        return None

    async def _fetch_job_results(self) -> None:
        """Fetch results for job_ids from CJM using api client."""
        async with ApiClient(config()) as client:
            page_reader = PageReader[PageResult, RawJobResult]()
            results_api = ResultsApi(client)
            pagination_handler = page_reader.get_all
            results_handler = results_api.read_results_by_job_id_results_job_job_id_get

            result_tasks = [
                pagination_handler(results_handler, job_id=circuit_data.job_id)
                for circuit_data in self.circuits_run_data
            ]
            result_items = await asyncio.gather(*result_tasks)
            job_ids_to_check = []

            for circuit_data, result_item in zip(self.circuits_run_data, result_items):
                circuit_data.results = None if not result_item else result_item[0]
                if circuit_data.results is None:
                    assert circuit_data.job_id is not None
                    job_ids_to_check.append(circuit_data.job_id)

            if job_ids_to_check:
                await self._fetch_failed_jobs_message(client, job_ids_to_check)

    async def _fetch_failed_jobs_message(self, api_client: ApiClient, job_ids_to_check: List[int]) -> None:
        """Fetch messages for failed jobs and update circuit data accordingly.

        Args:
            api_client: The API client for job communication.
            job_ids_to_check: List of job IDs that need to be inspected.
        """

        jobs_api = JobsApi(api_client)

        job_tasks = [jobs_api.read_job_jobs_id_get(id=_id) for _id in job_ids_to_check]

        jobs: List[Job] = await asyncio.gather(*job_tasks)

        failed_job_id_to_message = {
            job.id: job.message if job.status == QIJobStatus.FAILED else "No Results" for job in jobs
        }

        for circuit_data in self.circuits_run_data:

            if circuit_data.job_id in failed_job_id_to_message:
                circuit_data.system_message = failed_job_id_to_message[circuit_data.job_id]

    def status(self) -> JobStatus:
        """Return the status of the (batch)job, among the values of ``JobStatus``."""

        # mapping of QI2 BatchJobStatus to Qiskit JobStatus
        status_map = {
            BatchJobStatus.QUEUED: JobStatus.QUEUED,
            BatchJobStatus.RESERVED: JobStatus.QUEUED,
            BatchJobStatus.PLANNED: JobStatus.QUEUED,
            BatchJobStatus.RUNNING: JobStatus.RUNNING,
            BatchJobStatus.FINISHED: JobStatus.DONE,
        }

        batch_job = run_async(self._fetch_batchjob_status())
        return status_map[batch_job.status]

    async def _fetch_batchjob_status(self) -> BatchJob:
        async with ApiClient(config()) as api_client:
            api_instance = BatchJobsApi(api_client)

            page_reader = PageReader[PageBatchJob, BatchJob]()
            batch_job = await page_reader.get_single(api_instance.read_batch_jobs_batch_jobs_get, id=self.batch_job_id)
            if batch_job is None:
                raise RuntimeError(f"No (batch)job with id {self.batch_job_id}")

            return batch_job

    def serialize(self, file_path: Union[str, Path]) -> None:
        """Serialize job information in this class to a file.

        Uses Qiskit serialization to write circuits to a .qpy file, and includes
        backend and and batch_job information in the metadata so that we can recover
        the associated data later.

        Args:
            file_path: The path to the file where the job information will be stored.
        """
        if len(self.circuits_run_data) == 0:
            raise ValueError("No circuits to serialize")

        with open(file_path, "wb") as file:
            for circuit_data in self.circuits_run_data:
                circuit_data.circuit.metadata["job_id"] = circuit_data.job_id
                circuit_data.circuit.metadata["backend_type_name"] = self.backend().name
                circuit_data.circuit.metadata["backend_type_id"] = self.backend().id
                circuit_data.circuit.metadata["batch_job_id"] = self.batch_job_id

            qpy.dump([circuit_data.circuit for circuit_data in self.circuits_run_data], file)

    @classmethod
    def deserialize(cls, provider: BaseProvider, file_path: Union[str, Path]) -> "QIJob":
        """Recover a prior job from a file written by QIJob.serialize().

        Args:
            provider: Used to get the backend on which the original job ran.
            file_path: The path to the file where the job information is stored.
        """
        with open(file_path, "rb") as file:
            circuits = qpy.load(file)

            # Qiskit doesn't seem to allow serialization of an empty list of circuits
            assert len(circuits) > 0

            try:
                backend_name = cast(str, circuits[0].metadata["backend_type_name"])
                backend_id = cast(int, circuits[0].metadata["backend_type_id"])
                batch_job_id = cast(int, circuits[0].metadata["batch_job_id"])
            except KeyError:
                raise ValueError(f"Invalid file format: {file_path}")

            circuits = cast(list[QuantumCircuit], circuits)

            job = cls(circuits, provider.get_backend(backend_name, backend_id))
            job.batch_job_id = batch_job_id

            for circuit_data in job.circuits_run_data:
                circuit_data.job_id = circuit_data.circuit.metadata.get("job_id")

            return job

    @cache
    def result(self, wait_for_results: bool = True, timeout: float = 60.0) -> Result:
        """Return the results of the job."""
        if wait_for_results:
            self.wait_for_final_state(timeout=timeout)
        elif not self.done():
            raise RuntimeError(f"(Batch)Job status is {self.status()}.")
        run_async(self._fetch_job_results())
        return self._process_results()



_DEFAULT_QISKIT_TO_OPENSQUIRREL_MAPPING: dict[str, str] = {
    "id": "I",
    "h": "H",
    "x": "X",
    "y": "Y",
    "z": "Z",
    "s": "S",
    "sdg": "Sdag",
    "t": "T",
    "tdg": "Tdag",
    "rx": "Rx",
    "ry": "Ry",
    "rz": "Rz",
    "cx": "CNOT",
    "cz": "CZ",
    "cp": "CR",
    "swap": "SWAP",
    "measure": "measure",
    "reset": "reset",
    "barrier": "barrier",
    "delay": "wait",
    "ccx": "toffoli",
}


class InstructionMapping:
    def __init__(self, qiskit_to_os: dict[str, str] = _DEFAULT_QISKIT_TO_OPENSQUIRREL_MAPPING):
        self._QISKIT_TO_OPENSQUIRREL_MAPPING: dict[str, str] = qiskit_to_os
        # Uses lower case for keys to normalize inconsistent capitalization of backends
        self._OPENSQUIRREL_TO_QISKIT_MAPPING: dict[str, str] = {
            v.lower(): k for k, v in self._QISKIT_TO_OPENSQUIRREL_MAPPING.items()
        }

    def qiskit_to_opensquirrel(self, instruction: str) -> str:
        """Translate a Qiskit gate name to the equivalent opensquirrel gate name."""
        return self._QISKIT_TO_OPENSQUIRREL_MAPPING[instruction.lower()]

    def opensquirrel_to_qiskit(self, instruction: str) -> str:
        """Translate an opensquirrel gate name to the equivalent Qiskit gate name."""
        return self._OPENSQUIRREL_TO_QISKIT_MAPPING[instruction.lower()]

    def supported_opensquirrel_instructions(self) -> list[str]:
        """Return a list of all supported opensquirrel instructions."""
        return list(self._QISKIT_TO_OPENSQUIRREL_MAPPING.values())

    def supported_qiskit_instructions(self) -> list[str]:
        """Return a list of all supported Qiskit instructions."""
        return list(self._QISKIT_TO_OPENSQUIRREL_MAPPING.keys())



from pprint import PrettyPrinter
from typing import Any, List, Union

from compute_api_client import ApiClient, BackendStatus, BackendType, BackendTypesApi
from qi2_shared.client import config
from qiskit.circuit import QuantumCircuit
from qiskit.providers import BackendV2 as Backend
from qiskit.providers.options import Options
from qiskit.transpiler import CouplingMap, Target

from qiskit_quantuminspire.mapping.instruction_mapping import InstructionMapping
from qiskit_quantuminspire.qi_jobs import QIJob
from qiskit_quantuminspire.utils import is_coupling_map_complete, run_async

_IGNORED_GATES: set[str] = {
    # Prep not viewed as separate gates in Qiskit
    "prep_x",
    "prep_y",
    "prep_z",
    # Measure x and y not natively supported https://github.com/Qiskit/qiskit/issues/3967
    "measure_x",
    "measure_y",
    "measure_all",
    # Measure z is equivalent to measure
    "measure_z",
    # May be supportable through parameterized CPhaseGate.
    # For now, direct usage of CPhaseGate is required
    "crk",
    # No direct qiskit equivalent
    "x90",
    "mx90",
    "y90",
    "my90",
    # Qiskit assumes barrier support and does not include it in its standard gate mapping
    "barrier",
}


# Ignore type checking for QIBackend due to missing Qiskit type stubs,
# which causes the base class 'Backend' to be treated as 'Any'.
class QIBaseBackend(Backend):  # type: ignore[misc]
    _max_shots: int

    def __init__(self, backend_type: BackendType, mapping: InstructionMapping = InstructionMapping(), **kwargs: Any):
        super().__init__(name=backend_type.name, description=backend_type.description, **kwargs)
        self._id: int = backend_type.id

        self._max_shots: int = backend_type.max_number_of_shots

        # Construct options
        self._options = self._default_options()
        self.set_options(shots=backend_type.default_number_of_shots)

        if not backend_type.supports_raw_data:
            self._options.set_validator("memory", [False])

        # Determine supported gates
        opensquirrel_gates = {inst.lower() for inst in mapping.supported_opensquirrel_instructions()}
        available_gates = opensquirrel_gates - _IGNORED_GATES

        # Construct coupling map
        coupling_map = CouplingMap(backend_type.topology)
        coupling_map_complete = is_coupling_map_complete(coupling_map)

        if "toffoli" in available_gates and not coupling_map_complete:
            # "Toffoli gate not supported for non-complete topology
            available_gates.remove("toffoli")

        self._target = Target().from_configuration(
            basis_gates=[mapping.opensquirrel_to_qiskit(gate) for gate in available_gates],
            num_qubits=backend_type.nqubits,
            coupling_map=None if coupling_map_complete else coupling_map,
        )

    def __repr_pretty__(self, p: PrettyPrinter) -> None:
        p.pprint(f"QIBackend(name={self.name}, id={self.id})")

    def __repr__(self) -> str:
        module_name = self.__class__.__module__
        s = f"<{module_name}.{self.__class__.__name__} object at 0x{id(self):x} (name={self.name}, id={self.id})>"
        return s

    @classmethod
    def _default_options(cls) -> Options:
        """Only options defined here are supported by the backend.

        shots: int: Number of shots for the job.
        """
        options = Options(shots=1024, seed_simulator=None, memory=False)

        # Seed_simulator is included in options to enable use of BackendEstimatorV2 in Qiskit,
        # but is not actually supported by the backend so any other value than none raises an error.
        options.set_validator("seed_simulator", [None])

        options.set_validator("shots", int)
        options.set_validator("memory", bool)

        return options

    @property
    def target(self) -> Target:
        return self._target

    @property
    def max_shots(self) -> int:
        return self._max_shots

    @property
    def max_circuits(self) -> Union[int, None]:
        return None

    @property
    def id(self) -> int:
        return self._id


class QIBackend(QIBaseBackend):
    """A wrapper class for QuantumInspire backendtypes to integrate with Qiskit's Backend interface."""

    @property
    def status(self) -> BackendStatus:
        backend_type: BackendType = run_async(self._get_backend_type())
        return backend_type.status

    async def _get_backend_type(self) -> BackendType:
        async with ApiClient(config()) as client:
            backend_types_api = BackendTypesApi(client)
            return await backend_types_api.read_backend_type_backend_types_id_get(self._id)

    @property
    def available(self) -> bool:
        return bool(self.status != BackendStatus.OFFLINE)

    def run(self, run_input: Union[QuantumCircuit, List[QuantumCircuit]], **options: Any) -> QIJob:
        """Create and run a (batch)job on an QuantumInspire Backend.

        Args:
            run_input: A single or list of Qiskit QuantumCircuit objects or hybrid algorithms.

        Returns:
            QIJob: A reference to the batch job that was submitted.
        """
        if not self.available:
            raise RuntimeError(f"{self.name} is {self.status.value}, jobs can't be submitted")
        self.set_options(**options)
        job = QIJob(run_input=run_input, backend=self)
        job.submit()
        return job
